import { z } from 'zod';

// 令牌请求验证模式 - 授权码授权
export const authorizationCodeTokenRequestSchema = z.object({
    grant_type: z.literal('authorization_code'),
    code: z.string(),
    redirect_uri: z.string().url('必须是有效的URL'),
    client_id: z.string(),
    client_secret: z.string().optional(),
    code_verifier: z.string().optional(),
});

// 令牌请求验证模式 - 刷新令牌
export const refreshTokenRequestSchema = z.object({
    grant_type: z.literal('refresh_token'),
    refresh_token: z.string(),
    client_id: z.string(),
    client_secret: z.string().optional(),
    scope: z.string().optional(),
});

// 令牌请求验证模式 - 客户端凭证
export const clientCredentialsTokenRequestSchema = z.object({
    grant_type: z.literal('client_credentials'),
    client_id: z.string(),
    client_secret: z.string(),
    scope: z.string().optional(),
});

// 令牌请求验证模式 - 密码授权（可选支持）
export const passwordTokenRequestSchema = z.object({
    grant_type: z.literal('password'),
    username: z.string(),
    password: z.string(),
    client_id: z.string(),
    client_secret: z.string().optional(),
    scope: z.string().optional(),
});

// 合并的令牌请求验证模式
export const tokenRequestSchema = z.discriminatedUnion('grant_type', [
    authorizationCodeTokenRequestSchema,
    refreshTokenRequestSchema,
    clientCredentialsTokenRequestSchema,
    passwordTokenRequestSchema,
]);

// 令牌响应模式
export const tokenResponseSchema = z.object({
    access_token: z.string(),
    token_type: z.string().default('Bearer'),
    expires_in: z.number(),
    refresh_token: z.string().optional(),
    id_token: z.string().optional(),
    scope: z.string().optional(),
});

// 令牌内省请求模式
export const tokenIntrospectionRequestSchema = z.object({
    token: z.string(),
    token_type_hint: z.enum(['access_token', 'refresh_token']).optional(),
    client_id: z.string().optional(),
    client_secret: z.string().optional(),
});

// 令牌内省响应模式
export const tokenIntrospectionResponseSchema = z.object({
    active: z.boolean(),
    // 如果active为true，则提供以下信息
    scope: z.string().optional(),
    client_id: z.string().optional(),
    username: z.string().optional(),
    token_type: z.string().optional(),
    exp: z.number().optional(),
    iat: z.number().optional(),
    nbf: z.number().optional(),
    sub: z.string().optional(),
    aud: z.string().optional(),
    iss: z.string().optional(),
    jti: z.string().optional(),
});

// 令牌撤销请求模式
export const tokenRevocationSchema = z.object({
    token: z.string(),
    token_type_hint: z.enum(['access_token', 'refresh_token']).optional(),
    client_id: z.string(),
    client_secret: z.string().optional(),
});

// 错误响应模式
export const tokenErrorResponseSchema = z.object({
    error: z.enum([
        'invalid_request',
        'invalid_client',
        'invalid_grant',
        'unauthorized_client',
        'unsupported_grant_type',
        'invalid_scope'
    ]),
    error_description: z.string().optional(),
    error_uri: z.string().optional(),
});

// 类型定义
export type TokenRequest = z.infer<typeof tokenRequestSchema>;
export type TokenResponse = z.infer<typeof tokenResponseSchema>;
export type TokenIntrospectionRequest = z.infer<typeof tokenIntrospectionRequestSchema>;
export type TokenIntrospectionResponse = z.infer<typeof tokenIntrospectionResponseSchema>;
export type TokenRevocationRequest = z.infer<typeof tokenRevocationSchema>;
export type TokenErrorResponse = z.infer<typeof tokenErrorResponseSchema>; 